#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "kurganovlevy.h"
#include "row.h"
#include "boundary_conds.h"

double eos_gamma() { return EOS_GAMMA; }

/* 
 * forward_eos: given dens and ener, return pres and temp
 */
int forward_eos(int n, double *dens, double *ener, double *pres, double *temp) {

	int i;
	
	for (i=0; i<n; i++) {
		pres[i] = (EOS_GAMMA - 1.) * dens[i] * ener[i];
		temp[i] = ener[i] * ((EOS_GAMMA-1.)/EOS_GASCONST) * EOS_ABAR;
	}

	return 0;
}

int forward_eos_row(row_t *r) {
	int mustflip = 0;

	if (r->enerflag == FLAG_TOTALE) mustflip = 1;
	
	if (mustflip) ener_tot_to_internal(r);
	forward_eos(r->N, r->data[DENSVAR], r->data[ENERVAR], r->data[PRESVAR], r->data[TEMPVAR]);
	if (mustflip) ener_internal_to_tot(r);
	
	return 0;
}

int domain_forward_eos(domain_t *d) {

	int i,j;
	int nxt, nyt, nguard;
	
	nguard = d->Nguard;
	nxt = d->Nx+2*nguard;
	nyt = d->Ny+2*nguard;
	
	for (i=0; i<nxt; i++) {
		for (j=0; j<nyt; j++) {
			d->data[PRESVAR][i][j] = (EOS_GAMMA - 1.) * d->data[DENSVAR][i][j] * d->data[ENERVAR][i][j];
			d->data[TEMPVAR][i][j] = d->data[ENERVAR][i][j] * ((EOS_GAMMA-1.)/EOS_GASCONST) * EOS_ABAR;
		}
	}

	return 0;
}

int domain_timestep(domain_t *d, double *dt) {

	int i,j;
	int nx, ny, ng;
	double mindt  = 1e+19;
	double dx, dy;
	double tdt;
	double vx, vy, cs;
	
	ng = d->Nguard;
	nx = d->Nx;
	ny = d->Ny;

	dx = d->x[1] - d->x[0]; // assuming uniform grid
	dy = d->y[1] - d->y[0]; // assuming uniform grid
	
	for (i=ng; i<nx+ng; i++) {
		for (j=ng; j<ny+ng; j++) {
			vx = fabs(d->data[MOMXVAR][i][j]/d->data[DENSVAR][i][j]);
			vy = fabs(d->data[MOMYVAR][i][j]/d->data[DENSVAR][i][j]);
			cs = sqrt(EOS_GAMMA*d->data[PRESVAR][i][j]/d->data[DENSVAR][i][j]);
			tdt = dx/(cs + vx);
			if (dy/(cs+vy) < tdt) tdt = dy/(cs+vy);
			if (tdt < mindt) mindt = tdt;
		}
	}
	*dt = mindt;

	return 0;
}

/*
 * backward_dp_eos: given dens and presure, return temp and ener
 */

int backward_dp_eos(int n, double *dens, double *pres, double *temp, double *ener) {

	int i;
	
	for (i=0; i<n; i++) {
		ener[i] = pres[i]/((EOS_GAMMA-1.)*dens[i]);
		temp[i] = ener[i] * ((EOS_GAMMA -1.)/EOS_GASCONST) * EOS_ABAR;
	}

	return 0;
}


int backward_dp_eos_row(row_t *r) {
	int mustflip = 0;

	if (r->enerflag == FLAG_TOTALE) mustflip = 1;
	
	if (mustflip) ener_tot_to_internal(r);
	backward_dp_eos(r->N, r->data[DENSVAR], r->data[PRESVAR], r->data[TEMPVAR], r->data[ENERVAR]);
	if (mustflip) ener_internal_to_tot(r);

	return 0;
}

int domain_backward_dp_eos(domain_t *d) {

	int i,j;
	int nx, ny, nguard;
	
	nguard = d->Nguard;
	nx = d->Nx+2*nguard;
	ny = d->Ny+2*nguard;
	
	for (i=0; i<nx; i++) {
		for (j=0; j<ny; j++) {
			d->data[ENERVAR][i][j] = d->data[PRESVAR][i][j]/((EOS_GAMMA-1.)*d->data[DENSVAR][i][j]);
			d->data[TEMPVAR][i][j] = d->data[ENERVAR][i][j] * ((EOS_GAMMA -1.)/EOS_GASCONST) * EOS_ABAR;
		}
	}

	return 0;
}



/*
 * These routines implement 3rd order 1d-CWENO reconstructions assuming a uniform grid.
 */ 

int limit(int n, double *v, double min, double max) {
	int i;
	
	for (i=0; i<n; i++) {
		if (v[i] < min) v[i] = min;
	}
	for (i=0; i<n; i++) {
		if (v[i] > max) v[i] = max;
	}

	return 0;
}

int reconstruction(int n, double *u, double *x, double *wl, double *wc, double *wr, double p, double *a, double *b, double *c) {

	double eps = 1.e-9;
	double dx = x[1]-x[0];
	double isl, isc, isr;
	double alphal, alphac, alphar, alphatot;

	int i;

	for (i=1; i<n-1; i++) {
		isl = pow((u[i] - u[i-1]),2);
		isc = 13./3.*pow(( u[i+1] - 2.*u[i] + u[i-1] ),2) + 0.25*pow((u[i+1] - u[i-1]),2);
		isr = pow((u[i+1] - u[i]),2);

		alphal = 0.25  / pow(eps + isl,p);
		alphac = 0.50  / pow(eps + isc,p);
		alphar = 0.25  / pow(eps + isr,p);

		alphatot = alphar+alphac+alphal;

		wl[i] = alphal/alphatot;
		wr[i] = alphar/alphatot;
		wc[i] = alphac/alphatot;
	}

	a[0]   = u[0];   b[0]   = 0; c[0]   = 0;
	a[n-1] = u[n-1]; b[n-1] = 0; c[n-1] = 0;

	for (i=1; i<n-1; i++) {
		a[i] = u[i] - wc[i]/12.*(u[i+1] - 2*u[i] + u[i-1]);
		b[i] = (wr[i]*(u[i+1] - u[i]) + wc[i]/2.*(u[i+1] - u[i-1]) + wl[i]*(u[i] - u[i-1]))/dx;
		c[i] = 2.*wc[i]/(dx*dx) * ( u[i-1] - 2.*u[i] + u[i+1]);
		if (isnan(a[i])) {fprintf(stderr,"a[%d] = NaN; u[i] = %g, wc[i] = %g, u[i+1] = %g, u[i-1] = %g\n", i,  u[i], wc[i], u[i+1], u[i-1]);}
		if (isnan(b[i])) {fprintf(stderr,"b[%d] = NaN; u[i] = %g, wr[i] = %g, wc[i] = %g, wl[i] = %g, u[i+1] = %g, u[i-1] = %g\n", i,  u[i], wr[i], wc[i], wl[i], u[i+1], u[i-1]);}
		if (isnan(c[i])) {fprintf(stderr,"c[%d] = NaN; u[i] = %g, wr[i] = %g, wc[i] = %g, wl[i] = %g, u[i+1] = %g, u[i-1] = %g\n", i,  u[i], wr[i], wc[i], wl[i], u[i+1], u[i-1]);}
	}

	return 0;
}

/*
 * interpolate to interfaces from the right and left.
 */
int intvalues(int n, double *u, double *x, double *uhalfright, double *uhalfleft) {

	double *wl, *wc, *wr;
	double *a, *b, *c;
	double dx = x[1] - x[0];
	int i, status;

	wr = (double *)malloc(n * sizeof(double));
	wc = (double *)malloc(n * sizeof(double));
	wl = (double *)malloc(n * sizeof(double));
	if (wr == NULL || wc == NULL || wl == NULL) {
		fprintf(stderr,"intvalues: could not allocate 1d arrays; failing\n");
		return -2;
	}

	a = (double *)malloc(n * sizeof(double));
	b = (double *)malloc(n * sizeof(double));
	c = (double *)malloc(n * sizeof(double));
	if (a == NULL || b == NULL || c == NULL) {
		fprintf(stderr,"intvalues: could not allocate 1d arrays; failing\n");
		free(wr); free(wc); free(wl);
		return -2;
	}

	status = reconstruction(n, u, x, wl, wc, wr, 2., a, b, c);

	free(wr); free(wc); free(wl);

	/* interpolate to the ith boundary (between cell i-1 and i) from the right and left */
	for (i=0; i<n; i++) {
		uhalfright[i]   = a[i] - dx/2.*b[i] + dx*dx/8.*c[i];
		uhalfleft [i+1] = a[i] + dx/2.*b[i] + dx*dx/8.*c[i];
	}
	uhalfright[n]  = u[n-1];
	uhalfleft [0]  = u[0];
	
	free(a); free(b); free(c);

	return 0;
}
/*
 *  Implements a 1.5-d dimensionally split Kurganov-Levy hydro solver (eg, 
 *  carries around a passive tangential velocity).  To be used in a 
 *  Strang-split 2d solver.
 *
 *  Ref: http://arxiv.org/abs/math.NA/0002133
 */


/* define some floors */
#define KL_MINE 1.e-13
#define KL_MIND 1.e-13
#define KL_MAXE 1.e+29
#define KL_MAXD 1.e+29

double max(double x, double y) {
	return (x > y ? x : y);	
}


int calcflux(row_t *u, row_t *f) {
	
	int i, mustflip = 0;
	double vx;

	if (f->N < u->N) {
		printf("calcflux: sizes don't match.  Failing\n");	
		return -1;
	}
	
	if (u->enerflag != FLAG_TOTALE) mustflip = 1;
	if (mustflip) ener_internal_to_tot(u);
	for (i=0; i<u->N; i++) {
		vx = u->data[MOMXVAR][i]/u->data[DENSVAR][i];

		f->data[DENSVAR][i] = u->data[MOMXVAR][i];
		f->data[MOMXVAR][i] = u->data[MOMXVAR][i]*vx + u->data[PRESVAR][i];
		f->data[MOMYVAR][i] = u->data[MOMYVAR][i]*vx;
		f->data[ENERVAR][i] = (u->data[ENERVAR][i] + u->data[PRESVAR][i])*vx;
	}
	if (mustflip) ener_tot_to_internal(u);

	return 0;
}

int kl_semidiscrete_row(row_t *r, int nguard, double *ddensdt, double *dmomxdt, double *dmomydt, double *denerdt) {
                                      

	/* assume uniform grid */
        double dx = r->x[1]-r->x[0];
	double gamma;
	
	int i,n;

	row_t leftint;
	row_t rightint;
	row_t fluxleft, fluxright;
	double *csleft, *csright, *csint;

	int status;

	n = r->N;
	gamma = eos_gamma();

	/* there is one more interface than cell in a 1d array */
	status = init_row(&leftint,  n+1, r->x[0]-dx/2., r->x[n-1]+dx/2., FLAG_TOTALE);
	status+= init_row(&rightint, n+1, r->x[0]-dx/2., r->x[n-1]+dx/2., FLAG_TOTALE);
	status+= init_row(&fluxleft, n+1, r->x[0]-dx/2,  r->x[n-1]+dx/2., FLAG_TOTALE);
	status+= init_row(&fluxright,n+1, r->x[0]-dx/2,  r->x[n-1]+dx/2., FLAG_TOTALE);

	if (status != 0) {
		fprintf(stderr,"Could not create rows for kl_semidiscrete; failing.\n");
		return -1;
	}
	
	csleft  = (double *)malloc((n+1) * sizeof(double));
	csright = (double *)malloc((n+1) * sizeof(double));
	
	/* csint[i] is the sound speed at the interface between cell i-1 and i */
	csint   = (double *)malloc((n+1) * sizeof(double));

	if (csleft == NULL || csright == NULL || csint == NULL) {
		fprintf(stderr,"Could not malloc sound speed arrays; failing.\n");
		free_row(&leftint); free_row(&rightint); free_row(&fluxleft); free_row(&fluxright);
		return -1;
	}

	/* find interface values of primitive variables */
	/* interface i is between cell i-1 and i; interpolated from right and left */
	status = intvalues(n, r->data[DENSVAR], r->x, rightint.data[DENSVAR], leftint.data[DENSVAR]);
	status+= intvalues(n, r->data[MOMXVAR], r->x, rightint.data[MOMXVAR], leftint.data[MOMXVAR]);
	status+= intvalues(n, r->data[MOMYVAR], r->x, rightint.data[MOMYVAR], leftint.data[MOMYVAR]);
	status+= intvalues(n, r->data[ENERVAR], r->x, rightint.data[ENERVAR], leftint.data[ENERVAR]);
	if (status != 0) {
		fprintf(stderr,"kl_semidiscrete cound not interpolate values; failing.\n");
		return -2;
	}

	limit(n, rightint.data[DENSVAR], KL_MIND, KL_MAXD);
	limit(n, leftint .data[DENSVAR], KL_MIND, KL_MAXD);

	ener_tot_to_internal(&leftint);  limit(n, leftint.data[ENERVAR], KL_MINE, KL_MAXE);
	ener_tot_to_internal(&rightint); limit(n, rightint.data[ENERVAR], KL_MINE, KL_MAXE);

        forward_eos_row(&leftint);
        forward_eos_row(&rightint);

	/* find maximum signal speeds */
	for (i=0; i<n+1; i++) {
		csleft[i] = sqrt(gamma * leftint.data[PRESVAR][i] /leftint.data[DENSVAR][i]);
		csright[i]= sqrt(gamma * rightint.data[PRESVAR][i]/rightint.data[DENSVAR][i]);

		csleft[i] = csleft[i] + fabs(leftint .data[MOMXVAR][i]/leftint.data[DENSVAR][i]);
		csright[i]= csright[i]+ fabs(rightint.data[MOMXVAR][i]/rightint.data[DENSVAR][i]);
	}

	for (i=0; i<n+1; i++) {
		csint[i] = max(csright[i],csleft[i]);
	}

	calcflux(&leftint, &fluxleft);
	calcflux(&rightint,&fluxright);

	for (i=0; i<n-1; i++) {
		ddensdt[i] = -(fluxright.data[DENSVAR][i+1] + fluxleft.data[DENSVAR][i+1] - fluxright.data[DENSVAR][i] - fluxleft.data[DENSVAR][i])/(2.*dx) + csint[i+1]/(2.*dx)*(rightint.data[DENSVAR][i+1] - leftint.data[DENSVAR][i+1]) - csint[i]/(2.*dx) * (rightint.data[DENSVAR][i] - leftint.data[DENSVAR][i]);
		dmomxdt[i] = -(fluxright.data[MOMXVAR][i+1] + fluxleft.data[MOMXVAR][i+1] - fluxright.data[MOMXVAR][i] - fluxleft.data[MOMXVAR][i])/(2.*dx) + csint[i+1]/(2.*dx)*(rightint.data[MOMXVAR][i+1] - leftint.data[MOMXVAR][i+1]) - csint[i]/(2.*dx) * (rightint.data[MOMXVAR][i] - leftint.data[MOMXVAR][i]);
		dmomydt[i] = -(fluxright.data[MOMYVAR][i+1] + fluxleft.data[MOMYVAR][i+1] - fluxright.data[MOMYVAR][i] - fluxleft.data[MOMYVAR][i])/(2.*dx) + csint[i+1]/(2.*dx)*(rightint.data[MOMYVAR][i+1] - leftint.data[MOMYVAR][i+1]) - csint[i]/(2.*dx) * (rightint.data[MOMYVAR][i] - leftint.data[MOMYVAR][i]);
		denerdt[i] = -(fluxright.data[ENERVAR][i+1] + fluxleft.data[ENERVAR][i+1] - fluxright.data[ENERVAR][i] - fluxleft.data[ENERVAR][i])/(2.*dx) + csint[i+1]/(2.*dx)*(rightint.data[ENERVAR][i+1] - leftint.data[ENERVAR][i+1]) - csint[i]/(2.*dx) * (rightint.data[ENERVAR][i] - leftint.data[ENERVAR][i]);
	}

	/* don't update the guardcells */
	for (i=0; i<nguard; i++) {
		ddensdt[i] = 0.;
		ddensdt[n-nguard+i] = 0.;
		dmomxdt[i] = 0.;
		dmomxdt[n-nguard+i] = 0.;
		dmomydt[i] = 0.;
		dmomydt[n-nguard+i] = 0.;
		denerdt[i] = 0.;
		denerdt[n-nguard+i] = 0.;
	}


	free_row(&leftint);
	free_row(&rightint);
	free_row(&fluxleft);
	free_row(&fluxright);
	free(csleft); free(csright); free(csint);

	return 0;
}


/*
 * Does a 2nd order RK time integration of the hydro equations, using 
 * the semidiscrete routine to get an instantaneous dX/dt.
 *
 * Could do 3rd order RK but the Strang splitting makes us at best
 * 2nd order in time 
 */
int kl_row_rk2(row_t *r, double dt) {

	row_t rl;
	int n, status;
	int i;
	double *ddensdt, *dmomxdt, *dmomydt, *denerdt;

	n = r->N;

	ddensdt = (double *)malloc(n * sizeof(double));
	dmomxdt = (double *)malloc(n * sizeof(double));
	dmomydt = (double *)malloc(n * sizeof(double));
	denerdt = (double *)malloc(n * sizeof(double));

	if (ddensdt == NULL || dmomxdt == NULL || dmomydt == NULL || denerdt == 0) {
		fprintf(stderr,"kurganovlevy_row_rk2: cannot allocate arrays for derivs.  Failing\n");
		return -2;
	}

	/* eos of input data and ensure in valid range */
	ener_tot_to_internal(r);
	limit(n, r->data[ENERVAR], KL_MINE, KL_MAXE);
        forward_eos_row(r);
	ener_internal_to_tot(r);
	
	/* intermediate stage data needed for RK step */	
	status = init_row(&rl, n, r->x[0], r->x[n-1], FLAG_TOTALE);
	if (status != 0) {
		fprintf(stderr,"kurganovlevy_row_rk2; cannot initialize intermediate row.  Failing\n");
		free(ddensdt); free(dmomxdt); free(dmomydt); free(denerdt);
		return -2;
	}

	/* first stage of RK; get dX/dt at current timestep */
	kl_semidiscrete_row(r, KL_NGUARD, ddensdt, dmomxdt, dmomydt, denerdt);

	for (i=0; i<n; i++) {
		rl.data[DENSVAR][i] = r->data[DENSVAR][i] + dt*ddensdt[i];
		rl.data[MOMXVAR][i] = r->data[MOMXVAR][i] + dt*dmomxdt[i];
		rl.data[MOMYVAR][i] = r->data[MOMYVAR][i] + dt*dmomydt[i];
		rl.data[ENERVAR][i] = r->data[ENERVAR][i] + dt*denerdt[i];
	}

	limit(n, rl.data[DENSVAR], KL_MIND, KL_MAXD);
	ener_tot_to_internal(&rl);
	limit(n, rl.data[ENERVAR], KL_MINE, KL_MAXD);
	forward_eos_row(&rl);
	ener_internal_to_tot(&rl);

	/* predictor-corrector type step */
	kl_semidiscrete_row(&rl, KL_NGUARD, ddensdt, dmomxdt, dmomydt, denerdt);
	for (i=0; i<n; i++) {
		rl.data[DENSVAR][i] = 0.5*r->data[DENSVAR][i] + (1.-0.5)*(rl.data[DENSVAR][i] + dt*ddensdt[i]);
		rl.data[MOMXVAR][i] = 0.5*r->data[MOMXVAR][i] + (1.-0.5)*(rl.data[MOMXVAR][i] + dt*dmomxdt[i]);
		rl.data[MOMYVAR][i] = 0.5*r->data[MOMYVAR][i] + (1.-0.5)*(rl.data[MOMYVAR][i] + dt*dmomydt[i]);
		rl.data[ENERVAR][i] = 0.5*r->data[ENERVAR][i] + (1.-0.5)*(rl.data[ENERVAR][i] + dt*denerdt[i]);
	}

	limit(n, rl.data[DENSVAR], KL_MIND, KL_MAXD);
	ener_tot_to_internal(&rl);
	limit(n, rl.data[ENERVAR], KL_MINE, KL_MAXD);
	forward_eos_row(&rl);
	ener_internal_to_tot(&rl);

	/* copy new data back into old row */
	memcpy(r->data[DENSVAR], rl.data[DENSVAR], n*sizeof(double));
	memcpy(r->data[MOMXVAR], rl.data[MOMXVAR], n*sizeof(double));
	memcpy(r->data[MOMYVAR], rl.data[MOMYVAR], n*sizeof(double));
	memcpy(r->data[ENERVAR], rl.data[ENERVAR], n*sizeof(double));
	memcpy(r->data[PRESVAR], rl.data[PRESVAR], n*sizeof(double));
	memcpy(r->data[TEMPVAR], rl.data[TEMPVAR], n*sizeof(double));

	free_row(&rl);
	free(ddensdt);
	free(dmomxdt);
	free(dmomydt);
	free(denerdt);

	return 0;
}


int kl_timestep_xy(domain_t *d, int *bcs, double dt) {

	/* strang splitting; XYYX */
	int i;
	row_t row, col;
	int nr, nc, ng;

	nr = d->Ny;
	nc = d->Nx;
	ng = d->Nguard;

	init_row(&row, nc+2*ng, d->x[0], d->x[nc+2*ng-1], d->enerflag);
	init_row(&col, nr+2*ng, d->y[0], d->y[nr+2*ng-1], d->enerflag);
			
	domain_forward_eos(d);

	/* xsweep along all rows */
	for (i=ng; i<nr+ng; i++) {
		domain_extract_row(d, &row, i, XDIR);
		kl_row_rk2(&row, dt);
		domain_insert_row(d, &row, i, XDIR);
	}

	/* ysweep along all cols */
	for (i=ng; i<nc+ng; i++) {
		domain_extract_row(d, &col, i, YDIR);
		kl_row_rk2(&col, dt);
		domain_insert_row(d, &col, i, YDIR);
	}

	domain_forward_eos(d);

	free_row(&row);
	free_row(&col);

	return 0;
}

int kl_timestep_yx(domain_t *d, int *bcs, double dt) {

	/* strang splitting; XYYX */
	int i;
	row_t row, col;
	int nr, nc, ng;

	nr = d->Ny;
	nc = d->Nx;
	ng = d->Nguard;

	init_row(&row, nc+2*ng, d->x[0], d->x[nc+2*ng-1], d->enerflag);
	init_row(&col, nr+2*ng, d->y[0], d->y[nr+2*ng-1], d->enerflag);
			
	domain_forward_eos(d);

	/* ysweep along all cols */
	for (i=ng; i<nc+ng; i++) {
		domain_extract_row(d, &col, i, YDIR);
		kl_row_rk2(&col, dt);
		domain_insert_row(d, &col, i, YDIR);
	}

	/* xsweep along all rows */
	for (i=ng; i<nr+ng; i++) {
		domain_extract_row(d, &row, i, XDIR);
		kl_row_rk2(&row, dt);
		domain_insert_row(d, &row, i, XDIR);
	}

	domain_forward_eos(d);

	free_row(&row);
	free_row(&col);

	return 0;
}
